from django.contrib import admin
from django.shortcuts import render

from .models import ImpostorLog


@admin.register(ImpostorLog)
class ImpostorAdmin(admin.ModelAdmin):
    fields = ("impostor", "imposted_as", "logged_in", "impostor_ip")
    list_display = (
        "impostor",
        "imposted_as",
        "impostor_ip",
        "logged_in",
        "logged_out",
        "token",
    )
    list_editable = ()
    actions_on_top = False
    actions_on_bottom = False
    ordering = ("-logged_in", "impostor")
    readonly_fields = (
        "impostor",
        "imposted_as",
        "impostor_ip",
        "logged_in",
        "logged_out",
        "token",
    )
    search_fields = ("impostor__username", "imposted_as__username")

    def add_view(self, request, form_url="", extra_context=None):
        request.method = "GET"
        return super(ImpostorAdmin, self).add_view(request, form_url, extra_context)

    def change_view(self, request, object_id, form_url="", extra_context=None):
        request.method = "GET"
        return super(ImpostorAdmin, self).change_view(request, form_url, extra_context)

    def delete_view(self, request, object_id, extra_context=None):
        model = self.model
        opts = model._meta
        app_label = opts.app_label
        return render(
            request,
            template_name="delete_nono.html",
            context={"app_label": app_label, "opts": opts},
        )



# Generated by Django 3.0.7 on 2020-06-16 18:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("impostor", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="impostorlog",
            name="token",
            field=models.CharField(
                blank=True, db_index=True, max_length=36, verbose_name="Token"
            ),
        ),
    ]



# Generated by Django 3.0.7 on 2020-06-13 00:44

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="ImpostorLog",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "impostor_ip",
                    models.GenericIPAddressField(
                        blank=True, null=True, verbose_name="Impostor's IP address"
                    ),
                ),
                (
                    "logged_in",
                    models.DateTimeField(auto_now_add=True, verbose_name="Logged on"),
                ),
                (
                    "logged_out",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="Logged out"
                    ),
                ),
                (
                    "token",
                    models.CharField(
                        blank=True, db_index=True, max_length=32, verbose_name="Token"
                    ),
                ),
                (
                    "imposted_as",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="imposted_as",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Logged in as",
                    ),
                ),
                (
                    "impostor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="impostor",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Impostor",
                    ),
                ),
            ],
            options={
                "verbose_name": "Impostor log",
                "verbose_name_plural": "Impostor logs",
                "ordering": ("-logged_in", "impostor"),
            },
        ),
    ]



"""template tags for impostor."""

from django import template

from impostor.models import ImpostorLog

register = template.Library()


@register.simple_tag
def get_impersonated_as(request):
    """return the impersonated_as template tag to include into the main template.

    :param request:
    :return:
    """
    try:
        impersonated_as = ImpostorLog.objects.get(
            token=request.session["impostor_token"]
        )
    except (ImpostorLog.DoesNotExist, KeyError):
        impersonated_as = None

    return impersonated_as



import uuid

from django.contrib.auth import get_user_model
from django.db import models
from django.utils.translation import gettext_lazy as _


class ImpostorLog(models.Model):
    impostor = models.ForeignKey(
        get_user_model(),
        verbose_name=_("Impostor"),
        related_name="impostor",
        db_index=True,
        on_delete=models.CASCADE,
    )
    imposted_as = models.ForeignKey(
        get_user_model(),
        verbose_name=_("Logged in as"),
        related_name="imposted_as",
        db_index=True,
        on_delete=models.CASCADE,
    )
    impostor_ip = models.GenericIPAddressField(
        verbose_name=_("Impostor's IP address"), null=True, blank=True
    )
    logged_in = models.DateTimeField(verbose_name=_("Logged on"), auto_now_add=True)
    # These last two will come into play with Django 1.3+, but are here now for easier migration
    logged_out = models.DateTimeField(
        verbose_name=_("Logged out"), null=True, blank=True
    )
    token = models.CharField(
        verbose_name=_("Token"), max_length=36, blank=True, db_index=True
    )

    def save(self, *args, **kwargs):
        """
        Custom save method

        :param args:
        :param kwargs:
        :return:
        """
        if not self.token and self.impostor:
            self.token = str(uuid.uuid4())

        super(ImpostorLog, self).save(*args, **kwargs)

    class Meta:
        """
        Description of the Meta Class.
        """

        verbose_name = _("Impostor log")
        verbose_name_plural = _("Impostor logs")
        ordering = ("-logged_in", "impostor")

    def __str__(self):
        """
        str representation of the object.

        :return:
        """
        return "{} as {}".format(self.impostor, self.imposted_as)



import pytest
from django.contrib.auth import authenticate, get_user_model

from impostor.backend import AuthBackend
from impostor.forms import BigAuthenticationForm
from impostor.models import ImpostorLog
from impostor.templatetags.impostor_tags import get_impersonated_as

admin_username = "real_test_admin"
admin_pass = "admin_pass"
admin_email = "admin@email.com"

user_username = "real_test_user"
user_email = "real@mail.com"
user_pass = "user_pass"

user_with_supergroup_pass = "1234qwer"

fake_admin_username = "fake_test_admin"
fake_user_username = "fake_test_user"


@pytest.mark.django_db
class TestImpostorLogin:
    def test_login_user(self):
        """checks that a regular user can login normally through the next backend"""
        u = authenticate(username=user_username, password=user_pass)
        real_user = get_user_model().objects.get(username=user_username)
        assert u == real_user

    def test_login_admin(self):
        """checks that an admin user can login normally through the next backend"""
        u = authenticate(username=admin_username, password=admin_pass)
        real_admin = get_user_model().objects.get(username=admin_username)
        assert u == real_admin

    def test_login_admin_as_user(self):
        """checks that an admin user can impersonate a regular user via
        impostor backend and that the login get reflected into the
        ImpostorLog table"""
        assert ImpostorLog.objects.count() == 0

        composed_username = "{} as {}".format(admin_username, user_username)
        u = authenticate(username=composed_username, password=admin_pass)
        real_user = get_user_model().objects.get(username=user_username)
        assert u == real_user

        # Check if logs contain an entry now
        logs_entries = ImpostorLog.objects.all()
        assert len(logs_entries) == 1

        entry = logs_entries[0]
        # today = datetime.date.today()
        # lin = entry.logged_in
        assert entry.impostor.username == admin_username
        assert entry.imposted_as.username == user_username

        # assert (lin.year == today.year and lin.month == today.month and lin.day == today.day)
        assert entry.token and entry.token.strip() != ""

    def test_missing_authenticate_field(self):
        """checks that user cannot login if user USERNAME_FIELD/username
        or password field is missing"""
        composed_username = "{} as {}".format(admin_username, user_username)
        user = authenticate(
            random_username_field=composed_username, password=admin_pass
        )
        assert user is None

    def test_user_is_not_returned_if_admin_user_not_found(self):
        """checks that admin user doesn't exist case are properly
        handled out by backend"""
        composed_username = "{} as {}".format(fake_admin_username, user_username)
        user = authenticate(username=composed_username, password=admin_pass)
        assert user is None

    def test_user_is_not_returned_if_user_not_found(self):
        """checks that normal user doesn't exist case are properly
        handled out by backend"""
        composed_username = "{} as {}".format(admin_username, fake_user_username)
        user = authenticate(username=composed_username, password=admin_pass)
        assert user is None

    def test_form(self):
        """test custom login form"""
        initial = {"username": user_username, "password": user_pass}
        form = BigAuthenticationForm(data=initial)
        assert form.is_valid()
        assert form.cleaned_data["username"] == user_username
        assert form.cleaned_data["password"] == user_pass

        # Longer than contrib.auth default of 30 chars
        new_uname = "{} as {}".format(admin_username, user_username)
        initial = {"username": new_uname, "password": admin_pass}
        form = BigAuthenticationForm(data=initial)
        assert form.is_valid()
        assert form.cleaned_data["username"] == new_uname
        assert form.cleaned_data["password"] == admin_pass

        del initial["password"]
        form = BigAuthenticationForm(data=initial)
        assert not form.is_valid()

    @pytest.mark.parametrize(
        "first_user,password,impersonated_user,expected",
        [
            (
                pytest.lazy_fixture("real_admin"),
                admin_pass,
                pytest.lazy_fixture("real_user"),
                "ok",
            ),
            (
                pytest.lazy_fixture("real_admin"),
                admin_pass,
                pytest.lazy_fixture("real_admin"),
                "ok",
            ),
            (
                pytest.lazy_fixture("real_user"),
                user_pass,
                pytest.lazy_fixture("real_admin"),
                "ko",
            ),
            (
                pytest.lazy_fixture("real_user_with_supergroup"),
                user_with_supergroup_pass,
                pytest.lazy_fixture("real_user"),
                "ok",
            ),
            (
                pytest.lazy_fixture("real_user_with_supergroup"),
                user_with_supergroup_pass,
                pytest.lazy_fixture("real_admin"),
                "ko",
            ),
        ],
    )
    def test_impersonation(
        self, first_user, password, impersonated_user, expected, custom_settings, rf
    ):
        """
        check different use cases of impersonation

        :param first_user:
        :param password:
        :param impersonated_user:
        :param expected:
        :param custom_settings:
        :param rf:
        :return:
        """
        setattr(rf, "META", {})
        rf.META["HTTP_X_FORWARDED_FOR"] = "127.0.0.1,192.168.0.1"
        assert ImpostorLog.objects.count() == 0
        composed_username = "{} as {}".format(
            first_user.username, impersonated_user.username
        )
        authenticated_user = authenticate(
            request=rf, username=composed_username, password=password
        )
        if expected == "ok":
            assert authenticated_user == impersonated_user
            assert ImpostorLog.objects.count() == 1
            log = ImpostorLog.objects.first()
            assert log.impostor == first_user
            assert log.imposted_as == impersonated_user
            assert log.impostor_ip == "127.0.0.1"
        else:
            assert authenticated_user is None

    @pytest.mark.parametrize(
        "user_passed, user_expected",
        [
            (
                pytest.lazy_fixture("real_user_with_supergroup"),
                pytest.lazy_fixture("real_user_with_supergroup"),
            ),
            (None, None),
        ],
    )
    def test_get_user(self, user_passed, user_expected):
        """
        check get_user method
        :param user_passed:
        :param user_expected:
        :return:
        """
        try:
            user_id = user_passed.id
        except AttributeError:
            user_id = None
        result = AuthBackend.get_user(user_id)
        assert result == user_expected

    @pytest.mark.parametrize("existing_attr", [True, False])
    def test_impostor_group(self, custom_settings, existing_attr):
        """
        check impostor_group property
        :param custom_settings:
        :return:
        """
        if existing_attr:
            delattr(custom_settings, "IMPOSTOR_GROUP")
            assert AuthBackend().impostor_group is None
        else:
            assert AuthBackend().impostor_group is not None

    @pytest.mark.parametrize("in_session,expected", [(True, True), (False, False)])
    def test_impersonated_as_tag(self, real_admin, real_user, rf, in_session, expected):
        obj = ImpostorLog.objects.create(impostor=real_admin, imposted_as=real_user)
        setattr(rf, "session", {})
        if in_session:
            rf.session["impostor_token"] = obj.token
        result = get_impersonated_as(rf)
        if expected:
            assert result == obj
        else:
            assert result != obj

    def test_impostor_log_str(self, real_admin, real_user):
        obj = ImpostorLog.objects.create(impostor=real_admin, imposted_as=real_user)
        assert str(obj) == "{} as {}".format(real_admin.username, real_user.username)



"""Impostor."""
__version__ = "3.1.0"
__author__ = "Marko Samastur"
__email__ = "markos@gaivo.net"
__mantainer__ = "Andreu Vallbona"
__mantainer_email__ = "avallbona@gmail.com"
default_app_config = "impostor.apps.ImpostorConfig"



from django import forms
from django.contrib.auth.forms import AuthenticationForm
from django.utils.translation import gettext_lazy as _


class BigAuthenticationForm(AuthenticationForm):
    username = forms.CharField(label=_("Username"), max_length=70)



import codecs
import os
import re

from setuptools import find_packages, setup


def get_metadata(package, field):
    """
    Return package data as listed in `__{field}__` in `init.py`.
    """
    init_py = codecs.open(os.path.join(package, "__init__.py"), encoding="utf-8").read()
    return re.search(
        "^__{}__ = ['\"]([^'\"]+)['\"]".format(field), init_py, re.MULTILINE
    ).group(1)


def read(fname):
    """
    Read the README.md file

    :param fname:
    :return:
    """
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return codecs.open(file_path, encoding="utf-8").read()


setup(
    name="Impostor",
    version=get_metadata("impostor", "version"),
    url="https://github.com/avallbona/Impostor/",
    author=get_metadata("impostor", "author"),
    author_email=get_metadata("impostor", "email"),
    maintainer=get_metadata("impostor", "mantainer"),
    maintainer_email=get_metadata("impostor", "mantainer_email"),
    description="Staff can login as a different user.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    license="MIT License",
    platforms=["any"],
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Framework :: Django :: 4.1",
        "Framework :: Django :: 4.2",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Utilities",
    ],
    install_requires=[
        "Django>=2.0",
    ],
)



import pytest
from django.contrib.auth import authenticate, get_user_model
from django.test import override_settings

pwd = "admin_pass"


@pytest.mark.django_db
class TestCustomUser:
    @staticmethod
    def create_user(**kwargs):
        user = get_user_model().objects.create(**kwargs)
        return user

    def test_regular_user_model_with_username(self, real_admin, real_user, rf):
        """test regular case, USERNAME_FIELD = username"""
        composed_username = "{} as {}".format(real_admin.username, real_user.username)
        authenticated_user = authenticate(
            request=rf, username=composed_username, password=pwd
        )
        assert authenticated_user == real_user

    @override_settings(AUTH_USER_MODEL="testapp.CustomUser1")
    def test_custom_user_model_with_email(self, rf):
        """test case custom model user, USERNAME_FIELD = email"""

        real_admin = self.create_user(
            username="custom1username",
            email="custom1admin@email.com",
            is_superuser=True,
        )
        real_admin.set_password(pwd)
        real_admin.save()

        real_user = self.create_user(
            username="custom1_real_test_user",
            email="custom1_real@mail.com",
        )
        real_user.set_password("user_pass")
        real_user.save()

        composed_username = "{} as {}".format(real_admin.email, real_user.email)
        authenticated_user = authenticate(
            request=rf, email=composed_username, password=pwd
        )
        assert authenticated_user == real_user

    @override_settings(AUTH_USER_MODEL="testapp.CustomUser2")
    def test_custom_user_model_with_user_identifier(self, rf):
        """test case custom model user, USERNAME_FIELD = user_identifier"""

        real_admin = self.create_user(
            username="custom2admin",
            user_identifier="custom2adminidentifier",
            email="custom2admin@email.com",
            is_superuser=True,
        )
        real_admin.set_password(pwd)
        real_admin.save()

        real_user = self.create_user(
            username="custom2user",
            user_identifier="custom2useridentifier",
            email="custom2_real@mail.com",
        )
        real_user.set_password("user_pass")
        real_user.save()

        composed_username = "{} as {}".format(
            real_admin.user_identifier, real_user.user_identifier
        )
        authenticated_user = authenticate(
            request=rf, user_identifier=composed_username, password=pwd
        )
        assert authenticated_user == real_user



import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.utils.translation import gettext_lazy as _

from .models import ImpostorLog

log = logging.getLogger(__name__)


class ImpostorException(Exception):
    pass


class AuthBackend:
    supports_anonymous_user = False
    supports_object_permissions = False
    supports_inactive_user = False
    separator = " as "

    @property
    def impostor_group(self):
        try:
            imp_group, __ = Group.objects.get_or_create(name=settings.IMPOSTOR_GROUP)
        except (Group.DoesNotExist, AttributeError):
            imp_group = None
        return imp_group

    def user_in_impostor_group(self, user):
        """
        check is the user is in the impostor group

        :param user:
        :return:
        """
        return self.impostor_group and self.impostor_group in user.groups.all()

    @staticmethod
    def ip_address(request):
        """
        try to obtain the ip address from request

        :param request:
        :return:
        """
        try:
            ip_addr = request.META.get(
                "HTTP_X_FORWARDED_FOR",
                request.META.get("HTTP_X_REAL_IP", request.META.get("REMOTE_ADDR", "")),
            )
        except AttributeError:
            ip_addr = ""
        # if there are several ip addresses separated by comma like HTTP_X_FORWARDED_FOR returns,
        # take only the first one, which is the client's address
        if "," in ip_addr:
            ip_addr = ip_addr.split(",", 1)[0].strip()
        return ip_addr

    def is_user_allowed_to_impersonate(self, adm_obj):
        """
        verifies is the user who wants to impersonate another user can do it

        :param adm_obj:
        :return:
        """
        return adm_obj.is_superuser or self.user_in_impostor_group(adm_obj)

    @staticmethod
    def save_impostor_token_into_session(log_entry, request):
        """
        tries to save the impostor token into session

        :param log_entry:
        :param request:
        :return:
        """
        if not log_entry.token or not request:
            return
        try:
            request.session["impostor_token"] = log_entry.token
        except Exception as e:
            log.info(msg=str(e))

    def authenticate(self, request, username=None, password=None, **kwargs):
        auth_user = None
        User = get_user_model()

        try:
            if not username:
                username = kwargs.get(User.USERNAME_FIELD)
            if username is None or password is None:
                raise ImpostorException(
                    _("Both USERNAME_FIELD and password is required")
                )
            # if the split does not generate the admin and uuser
            # raise and exception and continues with the following backend
            try:
                admin, uuser = [
                    uname.strip() for uname in username.split(self.separator)
                ]
            except ValueError:
                raise ImpostorException(_("Regular login, moving to next auth backend"))

            # Check if admin exists, authenticates and is
            # allowed to impersonate another user
            try:
                adm_obj = User._default_manager.get_by_natural_key(admin)
            except User.DoesNotExist:
                # Run the default password hasher once to reduce the timing
                # difference between an existing and a nonexistent user.
                # (see https://code.djangoproject.com/ticket/20760)
                User().set_password(password)
                raise ImpostorException(_("Admin user is not active"))

            if self.is_user_allowed_to_impersonate(adm_obj) and adm_obj.check_password(
                password
            ):
                # get the user we want to impersonate
                try:
                    auth_user = User._default_manager.get_by_natural_key(uuser)
                except User.DoesNotExist:
                    # default password hasher is not required to run since user
                    # password is not checked
                    raise ImpostorException(_("User is not active"))

                # Superusers can only be impersonated by other superusers
                if auth_user.is_superuser and not adm_obj.is_superuser:
                    auth_user = None
                    raise ImpostorException(
                        _("Superuser can only be impersonated by a superuser.")
                    )

                # creates the impostor log entry
                log_entry = ImpostorLog.objects.create(
                    impostor=adm_obj,
                    imposted_as=auth_user,
                    impostor_ip=self.ip_address(request),
                )

                # save impostor_token into the session
                self.save_impostor_token_into_session(log_entry, request)

        except Exception as e:  # Nope. Do nothing and let other backends handle it.
            log.info(msg=str(e))
        return auth_user

    @staticmethod
    def get_user(user_id):
        User = get_user_model()
        try:
            return User._default_manager.get(pk=user_id)
        except User.DoesNotExist:
            pass
